import numpy as np
import torch
from scipy.spatial.distance import cdist
import ot
from geot.sinkhorn_loss import SinkhornLoss


class PartialOT:
    def __init__(
        self,
        cost_matrix: np.ndarray,
        penalty_waste="max",
        normalize_cost: bool = False,
        entropy_regularized: bool = False,
        spatiotemporal: bool = False,
    ):
        """
        Initialize unbalanced OT class with cost matrix
        Arguments:
            cost_matrix (np.ndarray): 2-dim numpy array with pairwise costs
                between locations
            penalty_waste (float or "max"): How much to penalize "waste vector",
                i.e. mass export and import. Either y_pred float value, or "max"
                corresponding to the maximum cost in cost_matrix
            normalize_cost (bool): Whether to normalize cost matrix by dividing
                by the maximum cost.
            entropy_regularized (bool): Set to True for using Sinkhorn loss,i.e.
                entropy-regularized OT. By default using Wasserstein distance.
            spatiotemporal (bool): Set to True to compute the error for spatio
                temporal data (across space and time)
        """
        self.entropy_regularized = entropy_regularized
        if penalty_waste == "max":
            penalty_waste = np.max(cost_matrix)

        # extend cost matrix to account for mass import / export
        clen, cwid = cost_matrix.shape
        extended_cost_matrix = np.zeros((clen + 1, cwid + 1))
        extended_cost_matrix[:clen, :cwid] = cost_matrix
        extended_cost_matrix[clen, :] = penalty_waste
        extended_cost_matrix[:, cwid] = penalty_waste

        if normalize_cost:
            extended_cost_matrix = extended_cost_matrix / np.max(
                extended_cost_matrix
            )
        if entropy_regularized:
            self.sinkhorn_object = SinkhornLoss(
                extended_cost_matrix,
                blur=0.1,
                reach=0.01,
                scaling=0.1,
                mode="unbalanced",
                spatiotemporal=spatiotemporal,
            )
        else:
            assert (
                not spatiotemporal
            ), "Can only use spatiotemporal with entropy regularized OT"
            self.cost_matrix = extended_cost_matrix

    def to_tensor(self, array):
        if isinstance(array, np.ndarray):
            array = torch.from_numpy(array)
        if array.dim() == 1:
            array = array.unsqueeze(0)
        return array

    def __call__(self, y_pred, y_true, return_matrix=False):
        """Compute OT error between y_pred and y_true

        Args:
            y_pred: array or tensor with predictions. Shape (batch_size, N)
            y_true: array or tensor with predictions. Shape (batch_size, N)
            return_matrix (bool, optional): Whether to output the OT matrix.
                Defaults to False.

        Returns:
            float: Optimal transport distance between the two distributions
        """
        if return_matrix:
            assert (
                not self.entropy_regularized
            ), "Cannot return matrix for Sinkhorn"
        y_pred, y_true = self.to_tensor(y_pred), self.to_tensor(y_true)
        # convert to arrays
        # compute mass that has to be imported or exported
        diff = (
            torch.sum(y_pred, dim=-1) - torch.sum(y_true, dim=-1)
        ).unsqueeze(-1)
        diff_pos = torch.relu(diff)
        diff_neg = torch.relu(diff * -1)

        # extend
        extended_pred = torch.cat((y_pred, diff_neg), dim=-1)
        extended_true = torch.cat((y_true, diff_pos), dim=-1)

        if self.entropy_regularized:
            return self.sinkhorn_object(extended_pred, extended_true)

        assert torch.all(y_pred >= 0) and torch.all(
            y_true >= 0
        ), "y_pred or y_true cannot be negative"
        # compute exact OT error with Wasserstein package
        assert (
            y_pred.size()[0] == 1
            and y_true.size()[0] == 1
            and y_pred.dim() == 2
        )
        extended_pred_np = (
            extended_pred.squeeze().detach().numpy().astype(float)
        )
        extended_true_np = (
            extended_true.squeeze().detach().numpy().astype(float)
        )
        # Note: extended_pred_np and extended_true_np already have the same sum
        # We still need this normalization to avoid numeric errors
        extended_pred_np = (
            extended_pred_np
            / np.sum(extended_pred_np)
            * np.sum(extended_true_np)
        )
        if return_matrix:
            transport_matrix = ot.emd(
                extended_pred_np, extended_true_np, self.cost_matrix
            )
            return transport_matrix

        cost = ot.emd2(extended_pred_np, extended_true_np, self.cost_matrix)
        return cost


def partial_ot_paired(
    cost_matrix: np.ndarray,
    y_pred: np.ndarray,
    y_true: np.ndarray,
    return_matrix: bool = False,
    **kwargs_partialot,
):
    """Compute OT error between y_pred and y_true

    Args:
        cost_matrix: 2-dim numpy array with pairwise costs between locations
        y_pred: array or tensor with predictions. Shape (batch_size, N)
        y_true: array or tensor with predictions. Shape (batch_size, N)
        return_matrix (bool, optional): Whether to output the OT matrix.
            Defaults to False.

    Returns:
        float: Optimal transport distance between the two distributions
    """
    pot_obj = PartialOT(cost_matrix, **kwargs_partialot)
    return pot_obj(
        torch.from_numpy(y_pred),
        torch.from_numpy(y_true),
        return_matrix=return_matrix,
    )


def partial_ot_unpaired(
    locations_pred: np.ndarray,
    locations_gt: np.ndarray,
    cost_matrix=None,
    import_location=np.array([0, 0]),
    penalty_waste=0,
    return_matrix=False,
):
    """_summary_

    Args:
        locations_pred (np.ndarray): Predicted locations (array of shape (N, d))
        locations_gt (np.ndarray): Ground true locations (array of shape (M, d))
        cost_matrix (np.ndarray, optional): Pairwise costs between locations.
            If None, Euclidean distances are used. Defaults to None.
        import_location (np.ndarray, optional): Location for importing and
            exporting mass. Can be set by appliction, e.g. bike sharing
            distribution center Defaults to np.array([0, 0]).
        penalty_waste (float or "max"): How much to penalize "waste vector",
                i.e. mass export and import. Either y_pred float value, or "max"
                corresponding to the maximum cost in cost_matrix. Defaults to 0.
        return_matrix (bool, optional): Whether to output the OT matrix.
            Defaults to False.

    Returns:
        float: Optimal transport distance between the two distributions
    """
    # extend the shorter one with import / export coords
    len_diff = len(locations_pred) - len(locations_gt)
    fill_vector = np.expand_dims(import_location, 0).repeat(len_diff, axis=0)
    if len_diff < 0:
        locations_pred_ext = np.concatenate([locations_pred, fill_vector])
        locations_gt_ext = locations_gt
    else:
        locations_gt_ext = np.concatenate([locations_gt, fill_vector])
        locations_pred_ext = locations_pred

    # at each location is y_pred mass of 1
    new_len = len(locations_gt_ext)
    weights_pred = np.ones(new_len) / new_len
    weights_gt = np.ones(new_len) / new_len

    # compute cost matrix
    if cost_matrix is None:
        cost_matrix = cdist(locations_pred_ext, locations_gt_ext)
        if len_diff < 0:
            cost_matrix[-len_diff:, :] = penalty_waste
        else:
            cost_matrix[:, -len_diff:] = penalty_waste

    transport_matrix = ot.emd(weights_pred, weights_gt, cost_matrix)
    ot_distance = np.sum(transport_matrix * cost_matrix)

    if return_matrix:
        return transport_matrix

    else:
        return ot_distance



import geomloss
import torch
import numpy as np
from torch.nn import MSELoss

device = "cuda" if torch.cuda.is_available() else "cpu"
NONZERO_FACTOR = 1e-5


class SinkhornLoss:
    def __init__(
        self,
        cost_matrix,
        normalize_cost=True,
        spatiotemporal=False,
        blur=0.1,
        reach=0.01,
        scaling=0.1,
        mode="unbalanced",
        **sinkhorn_kwargs,
    ):
        """Initialize Sinkhorn loss to train NN with OT loss

        Args:
            cost_matrix (np.ndarray): 2-dim numpy array with pairwise costs
                between locations
            normalize_cost (bool): Whether to normalize cost matrix by dividing
                by the maximum cost.
            spatiotemporal (bool): Set to True to compute the error for spatio-
                temporal data (across space and time)
            blur (float, optional): typical scale associated to the temperature.
                Defaults to 0.1. See
                https://www.kernel-operations.io/geomloss/api/pytorch-api.html.
            reach (float, optional): specifies the typical scale associated to
                the constraint strength. Defaults to 0.01. See
                https://www.kernel-operations.io/geomloss/api/pytorch-api.html.
            scaling (float, optional): _description_. Defaults to 0.1.
                https://www.kernel-operations.io/geomloss/api/pytorch-api.html.
            mode (str, optional): How prediction and gt are normalized.
                if mode=unbalanced: pred and gt have different mass
                if mode=balancedSoftmax: pred is softmaxed, gt is normalized
                if mode=balanced: pred and gt are both normalized to sum 1
                Defaults to "unbalanced".
        """
        assert mode in ["unbalanced", "balancedSoftmax", "balanced"]
        self.mode = mode
        self.spatiotemporal = spatiotemporal
        # adapt cost matrix type and size
        if isinstance(cost_matrix, np.ndarray):
            cost_matrix = torch.from_numpy(cost_matrix)
        # normalize to values betwen 0 and 1
        if normalize_cost:
            cost_matrix = cost_matrix / torch.max(cost_matrix)
        if cost_matrix.dim() != 3:
            if cost_matrix.dim() != 2:
                raise ValueError(
                    "cost matrix cost_matrix must have 2 or 3 dimensions"
                )
            cost_matrix = cost_matrix.unsqueeze(0)

        # cost matrics and locs both need a static representation and are
        # modified later to match the batch size
        self.cost_matrix = cost_matrix.to(device)
        self.cost_matrix_original = self.cost_matrix.clone()

        # introduce dummy weights since we assume fixed locations
        self.dummy_weights_alpha = torch.tensor(
            [[[i] for i in range(cost_matrix.size()[-2])]]
        ).float()
        self.dummy_weights_beta = torch.tensor(
            [[[i] for i in range(cost_matrix.size()[-1])]]
        ).float()
        self.dummy_weights_a = self.dummy_weights_alpha.clone()
        self.dummy_weights_b = self.dummy_weights_beta.clone()

        # sinkhorn loss
        self.loss_object = geomloss.SamplesLoss(
            loss="sinkhorn",
            cost=self.get_cost,
            backend="tensorized",
            debias=True,
            blur=blur,
            reach=reach,
            scaling=scaling,
            **sinkhorn_kwargs,
        )

    def get_cost(self, a, b):
        return self.cost_matrix

    def adapt_to_batchsize(self, batch_size):
        if self.cost_matrix.size()[0] != batch_size:
            self.cost_matrix = self.cost_matrix_original.repeat(
                (batch_size, 1, 1)
            )
            self.dummy_weights_a = self.dummy_weights_alpha.repeat(
                (batch_size, 1, 1)
            )
            self.dummy_weights_b = self.dummy_weights_beta.repeat(
                (batch_size, 1, 1)
            )

    def __call__(self, a_in, b_in):
        """a_in: predictions, b_in: targets"""

        # 1) Normalize dependent on the OT mode (balanced / unbalanced)
        b_in = b_in + NONZERO_FACTOR  # to prevent that all gt are zero
        if self.mode == "balanced":
            a = a_in / torch.unsqueeze(torch.sum(a_in, dim=-1), -1)
            b = b_in / torch.unsqueeze(torch.sum(b_in, dim=-1), -1)
        elif self.mode == "balancedSoftmax":
            # this yields a spearman correlation of 0.74
            a = (a_in * 2.71828).softmax(dim=-1)
            b = b_in / torch.unsqueeze(torch.sum(b_in, dim=-1), -1)
        else:
            # TODO: Any other possibility to do relu without getting all zeros?
            a = torch.relu(a_in) + NONZERO_FACTOR
            b = b_in

        # 2) flatten one axis -> either for spatiotemporal OT or treating the
        # temporal axis as batch
        batch_size = a.size()[0]
        if self.spatiotemporal:
            # flatten space-time axes
            a = a.reshape((batch_size, -1))
            b = b.reshape((batch_size, -1))
        elif a.dim() == 3:
            # if we have to flatten at all, flatten time over the batch size
            steps_ahead = a.size()[1]
            a = a.reshape((batch_size * steps_ahead, -1))
            b = b.reshape((batch_size * steps_ahead, -1))
            batch_size = batch_size * steps_ahead

        # 3) Adapt cost matrix size to the batch size
        self.adapt_to_batchsize(batch_size)

        # 4) Normalize again if spatiotemporal (over the space-time axis)
        # such that it overall sums up to 1
        if self.spatiotemporal and self.mode != "unbalanced":
            a = a / torch.unsqueeze(torch.sum(a, dim=-1), -1)
            b = b / torch.unsqueeze(torch.sum(b, dim=-1), -1)

        loss = self.loss_object(
            a, self.dummy_weights_a, b, self.dummy_weights_b
        )
        return torch.sum(loss)


class CombinedLoss:
    def __init__(
        self, cost_matrix, mode="balancedSoftmax", spatiotemporal=False
    ) -> None:
        self.standard_mse = MSELoss()
        if spatiotemporal:
            self.sinkhorn_error = SinkhornLoss(
                cost_matrix, mode=mode, spatiotemporal=True
            )
            self.dist_weight = 100
        else:
            self.sinkhorn_error = SinkhornLoss(cost_matrix, mode=mode)
            self.dist_weight = 10

    def __call__(self, a_in, b_in):
        # compute the error between the mean of predicted and mean of gt demand
        # this is the overall demand per timestep per batch
        # total_mse = (torch.mean(a_in, dim=-1) - torch.mean(b_in, dim=-1)) ** 2
        # take the average of the demand divergence over batch & timestep
        # mse_loss = torch.mean(total_mse)
        mse_loss = self.standard_mse(a_in, b_in)
        # mse_loss = self.standard_mse(a_in, b_in)
        sink_loss = self.sinkhorn_error(a_in, b_in)
        # for checking calibration of weighting
        # print(mse_loss, self.dist_weight * sink_loss)
        return mse_loss + self.dist_weight * sink_loss


def sinkhorn_loss_from_numpy(
    a,
    b,
    cost_matrix,
    mode="unbalanced",
    sinkhorn_kwargs={},
    loss_class=SinkhornLoss,
):
    a = torch.tensor(a.tolist()).float()
    b = torch.tensor(b.tolist()).float()
    # cost_matrix = torch.tensor([cost_matrix])
    # # Testing for the case where multiple steps ahead are predicted
    # a = a.unsqueeze(1).repeat(1, 3, 1)
    # b = b.unsqueeze(1).repeat(1, 3, 1)
    # print("Before initializing", cost_matrix.shape, a.size(), b.size())
    loss = loss_class(cost_matrix, mode=mode, **sinkhorn_kwargs)
    return loss(a, b)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

plt.rcParams.update({"font.size": 13})


def plot_cost_matrix(cost_matrix, title="Cost matrix", label="Cost"):
    """
    Plot the cost matrix as a heatmap

    Args:
        cost_matrix (np.array): 2D array of shape (M, N) with the pairwise costs between M and N locations
    """
    assert cost_matrix.ndim == 2, "Cost matrix must be 2D"
    m, n = cost_matrix.shape
    plt.imshow(cost_matrix)
    plt.xlabel(f"Locations 1-{m}")
    plt.ylabel(f"Locations 1-{n}")
    plt.colorbar(label=label)
    plt.xticks([])
    plt.yticks([])
    plt.title(title)
    plt.show()


def plot_predictions_and_ground_truth(locations, predictions, observations):
    """
    Plot the predictions, observations and residuals as scatter plots

    Args:
        locations (np.array): 2D array of shape (N, 2) with the locations of the predictions and observations
        predictions (np.array): 1D array of shape (N,) with the predictions
        observations (np.array): 1D array of shape (N,) with the observations
    """
    # plot predictions, observations and residuals
    plt.figure(figsize=(15, 4))
    plt.subplot(1, 3, 1)
    plt.scatter(locations[:, 0], locations[:, 1], c=observations)
    plt.colorbar(label="observations")
    plt.xlabel("X coordinate")
    plt.ylabel("Y coordinate")
    plt.title("Observations")
    plt.subplot(1, 3, 2)
    plt.scatter(locations[:, 0], locations[:, 1], c=predictions)
    plt.colorbar(label="prediction")
    plt.xlabel("X coordinate")
    plt.ylabel("Y coordinate")
    plt.title("Predictions")
    plt.subplot(1, 3, 3)
    plt.scatter(locations[:, 0], locations[:, 1], c=predictions - observations)
    plt.colorbar(label="residuals")
    plt.xlabel("X coordinate")
    plt.ylabel("Y coordinate")
    plt.title("Residuals (predictions - observations)")
    plt.tight_layout()
    plt.show()


def plot_paired_transport_matrix(
    locations, predictions, observations, ot_matrix
):
    """
    Plot the transport matrix as arrows between predicted and true spatial
    distribution.

    Args:
        locations (np.array): 2D array of shape (N, 2), the locations
            (same locations for gt and preds)
        predictions (np.array): array of shape (N), the predictions
        observations (np.array): array of shape (N), the observations
        transport_matrix (np.array): 2D array of shape (N+1, N+1), the optimal
            transport matrix between the predicted and true spatial distribution
            (+1 for the waste vector)
    """
    head_width = 0.02 * np.mean(
        np.linalg.norm(
            locations - locations[np.random.permutation(len(locations))],
            axis=1,
        )
    )
    # get lists of indices for start and end stations
    start_station_id, end_station_id = np.where(ot_matrix[:-1, :-1] > 0)
    start_coords, end_coords = (
        locations[start_station_id],
        locations[end_station_id],
    )

    errors = np.abs(predictions - observations)
    residuals = predictions - observations

    def get_col(val):
        # one col for correct, one for over, one for under estimation
        if val < 0:
            return "purple"
        elif val == 0:
            return "green"
        else:
            return "orange"

    cols = [get_col(val) for val in residuals]

    fig, ax = plt.subplots(figsize=(10, 5))
    for (x1, y1), (x2, y2) in zip(start_coords, end_coords):
        plt.arrow(
            x1,
            y1,
            x2 - x1,
            y2 - y1,
            head_width=head_width,
            length_includes_head=True,
            overhang=1,
            alpha=0.5,
        )
    plt.scatter(
        locations[:, 0], locations[:, 1], s=errors * 100 + 10, c=cols, alpha=0.5
    )
    plt.xlabel("X coordinate")
    plt.ylabel("Y coordinate")
    legend_elements = [
        Line2D(
            [0],
            [0],
            marker="o",
            color="green",
            lw=0,
            markerfacecolor="g",
            markersize=7,
            label="Correct",
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            color="orange",
            lw=0,
            label="Prediction > GT",
            markerfacecolor="orange",
            markersize=7,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            color="purple",
            lw=0,
            label="GT > Prediction",
            markerfacecolor="purple",
            markersize=7,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            color="black",
            lw=0,
            label="High error",
            markerfacecolor="black",
            markersize=15,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            color="black",
            lw=0,
            label="Low error",
            markerfacecolor="black",
            markersize=3,
        ),
        Line2D([0], [0], color="black", lw=2, label="Transported mass"),
    ]

    # Create the figure
    ax.legend(handles=legend_elements, bbox_to_anchor=(1, 1))
    plt.tight_layout()
    plt.show()


def plot_unpaired_transport_matrix(
    predicted_locations, true_locations, transport_matrix
):
    """Plot the transport matrix as arrows between predicted and true locations

    Args:
        predicted_locations (np.array): 2D array of shape (N, 2), the predicted locations
        true_locations (np.array): 2D array of shape (M, 2), the true locations
        transport_matrix (np.array): 2D array of shape (N, M), the optimal transport matrix between the predicted and true locations
    """
    plt.figure(figsize=(5, 3))
    # plot points
    plt.scatter(
        predicted_locations[:, 0],
        predicted_locations[:, 1],
        label="predicted locations",
    )
    plt.scatter(
        true_locations[:, 0], true_locations[:, 1], label="true locations"
    )
    # compute suitable head with for the errors based on the scale
    head_with = 0.02 * np.mean(
        np.linalg.norm(
            true_locations
            - true_locations[np.random.permutation(len(true_locations))],
            axis=1,
        )
    )

    # plot arrows
    for i, (x1, y1) in enumerate(predicted_locations):
        for j, (x2, y2) in enumerate(true_locations):
            if (
                (i != j)
                and (transport_matrix[i, j] > 0)
                and not (x1 == x2 and y1 == y2)
            ):
                plt.arrow(
                    x1,
                    y1,
                    x2 - x1,
                    y2 - y1,
                    head_width=head_with,
                    length_includes_head=True,
                    overhang=1,
                )
    plt.xlabel("X coordinate")
    plt.ylabel("Y coordinate")
    plt.xticks([])
    plt.yticks([])
    plt.legend(loc="upper left", framealpha=1)
    plt.tight_layout()
    plt.show()



# test with differet datatypes
# test that it is the same as balanced OT for balanced data
# test with different input types and shapes
import numpy as np
import torch
from geot.partialot import PartialOT, partial_ot_paired

test_cdist = np.array(
    [
        [0.0, 0.9166617229649182, 0.8011636143804466, 1.0],
        [0.9166617229649182, 0.0, 0.2901671214052399, 0.5131642591866252],
        [0.8011636143804466, 0.2901671214052399, 0.0, 0.28166962442054133],
        [1.0, 0.5131642591866252, 0.28166962442054133, 0.0],
    ]
)
test_pred, test_gt = (np.array([[1, 2, 3, 4]]), np.array([[1, 3, 2, 4]]))


class TestPartialOT:
    """Test main class for partial OR"""

    def test_zero_for_same_mass(self):
        """Test OT error being zero for same distributions"""
        ot_obj = PartialOT(test_cdist, entropy_regularized=False)
        ot_error = ot_obj(
            torch.tensor([[1, 2, 3, 4]]),
            torch.tensor([[1, 2, 3, 4]]),
        )
        assert ot_error == 0

    def test_value_correct(self):
        """Test OT error being correct for hard-coded example"""
        ot_obj = PartialOT(
            test_cdist,
            entropy_regularized=False,
            penalty_waste=0,
        )
        ot_error = ot_obj(test_pred, test_gt)
        # compute with function
        function_computation = partial_ot_paired(
            test_cdist,
            test_pred,
            test_gt,
            penalty_waste=0,
        )
        # compute via matrix
        ot_matrix = partial_ot_paired(
            test_cdist,
            test_pred,
            test_gt,
            penalty_waste=0,
            return_matrix=True,
        )
        assert np.isclose(ot_error, 0.29016712)
        assert np.isclose(ot_error, np.sum(ot_matrix[:-1, :-1] * test_cdist))
        assert np.isclose(ot_error, function_computation)
